@@ -16,26 +16,27 @@
 module Network.Wai.Middleware.Jsonp (jsonp) where
 
 import Network.Wai
+import Network.Wai.Internal
 import Data.ByteString (ByteString)
 import qualified Data.ByteString.Char8 as B8
-import Blaze.ByteString.Builder (copyByteString)
+import Blaze.ByteString.Builder (Builder, copyByteString)
 import Blaze.ByteString.Builder.Char8 (fromChar)
 import Data.Monoid (mappend)
 import Control.Monad (join)
 import Data.Maybe (fromMaybe)
 import qualified Data.ByteString as S
-import qualified Data.Conduit as C
-import qualified Data.Conduit.List as CL
+import Data.CaseInsensitive (CI)
+import Network.HTTP.Types (Status)
 
 -- | Wrap json responses in a jsonp callback.
 --
 -- Basically, if the user requested a \"text\/javascript\" and supplied a
 -- \"callback\" GET parameter, ask the application for an
--- \"application/json\" response, then convern that into a JSONP response,
+-- \"application/json\" response, then convert that into a JSONP response,
 -- having a content type of \"text\/javascript\" and calling the specified
 -- callback function.
 jsonp :: Middleware
-jsonp app env = do
+jsonp app env sendResponse = do
     let accept = fromMaybe B8.empty $ lookup "Accept" $ requestHeaders env
     let callback :: Maybe B8.ByteString
         callback =
@@ -50,25 +51,25 @@ jsonp app env = do
                                            "application/json"
                                            $ requestHeaders env
                         }
-    res <- app env'
-    case callback of
-        Nothing -> return res
-        Just c -> go c res
+    app env' $ \res ->
+        case callback of
+            Nothing -> sendResponse res
+            Just c -> go c res
   where
     go c r@(ResponseBuilder s hs b) =
-        case checkJSON hs of
-            Nothing -> return r
-            Just hs' -> return $ ResponseBuilder s hs' $
+        sendResponse $ case checkJSON hs of
+            Nothing -> r
+            Just hs' -> responseBuilder s hs' $
                 copyByteString c
                 `mappend` fromChar '('
                 `mappend` b
                 `mappend` fromChar ')'
     go c r =
         case checkJSON hs of
-            Just hs' -> addCallback c s hs' b
-            Nothing -> return r
+            Just hs' -> addCallback c s hs' wb
+            Nothing -> sendResponse r
       where
-        (s, hs, b) = responseSource r
+        (s, hs, wb) = responseToStream r
 
     checkJSON hs =
         case lookup "Content-Type" hs of
@@ -78,11 +79,11 @@ jsonp app env = do
             _ -> Nothing
     fixHeaders = changeVal "Content-Type" "text/javascript"
 
-    addCallback cb s hs b =
-        return $ ResponseSource s hs $
-            CL.sourceList [C.Chunk $ copyByteString cb `mappend` fromChar '(']
-            `mappend` b
-            `mappend` CL.sourceList [C.Chunk $ fromChar ')']
+    addCallback cb s hs wb =
+        wb $ \body -> sendResponse $ responseStream s hs $ \sendChunk flush -> do
+            sendChunk $ copyByteString cb `mappend` fromChar '('
+            body sendChunk flush
+            sendChunk $ fromChar ')'
 
 changeVal :: Eq a
           => a
