@@ -1,6 +1,8 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards #-}
-{-# LANGUAGE CPP #-}
+
+-- NOTE: Due to https://github.com/yesodweb/wai/issues/192, this module should
+-- not use CPP.
 module Network.Wai.Middleware.RequestLogger
     ( -- * Basic stdout logging
       logStdout
@@ -22,40 +24,37 @@ import System.IO (Handle, stdout)
 import qualified Data.ByteString as BS
 import Data.ByteString.Char8 (pack, unpack)
 import Control.Monad.IO.Class (liftIO)
-import Network.Wai (Request(..), Middleware, responseStatus, Response)
+import Network.Wai (Request(..), Middleware, responseStatus, Response, responseHeaders)
 import System.Log.FastLogger
 import Network.HTTP.Types as H
 import Data.Maybe (fromMaybe)
+import Data.Monoid (mconcat)
 
 import Network.Wai.Parse (sinkRequestBody, lbsBackEnd, fileName, Param, File, getRequestBodyType)
 import qualified Data.ByteString.Lazy as LBS
-
-import qualified Data.Conduit as C
-import qualified Data.Conduit.List as CL
+import qualified Data.ByteString.Char8 as S8
 
 import System.Console.ANSI
-import Data.IORef
+import Data.IORef.Lifted
 import System.IO.Unsafe
+import Control.Monad (unless)
+import Network.Wai.Internal (Response (ResponseRaw))
 
-import Data.Default (Default (def))
-import Network.Wai.Logger.Format (apacheFormat, IPAddrSource (..))
-#if MIN_VERSION_fast_logger(0,3,0)
-import System.Date.Cache (ondemandDateCacher)
-#else
-import System.Log.FastLogger.Date (getDate, dateInit, ZonedDate)
-#endif
+import Data.Default.Class (Default (def))
+import Network.Wai.Logger
+import Network.Wai.Middleware.RequestLogger.Internal
 
 data OutputFormat = Apache IPAddrSource
                   | Detailed Bool -- ^ use colors?
                   | CustomOutputFormat OutputFormatter
 
-type OutputFormatter = ZonedDate -> Request -> Status -> Maybe Integer -> [LogStr]
+type OutputFormatter = ZonedDate -> Request -> Status -> Maybe Integer -> LogStr
 
 data Destination = Handle Handle
-                 | Logger Logger
+                 | Logger LoggerSet
                  | Callback Callback
 
-type Callback = [LogStr] -> IO ()
+type Callback = LogStr -> IO ()
 
 data RequestLoggerSettings = RequestLoggerSettings
     {
@@ -78,42 +77,39 @@ instance Default RequestLoggerSettings where
 
 mkRequestLogger :: RequestLoggerSettings -> IO Middleware
 mkRequestLogger RequestLoggerSettings{..} = do
-    (callback, mgetdate) <-
-        case destination of
-            Handle h -> fmap fromLogger $ mkLogger autoFlush h
-            Logger l -> return $ fromLogger l
-            Callback c -> return (c, Nothing)
+    let (callback, flusher) =
+            case destination of
+                Handle h -> (BS.hPutStr h . logToByteString, return ())
+                Logger l -> (pushLogStr l, flushLogStr l)
+                Callback c -> (c, return ())
     case outputFormat of
         Apache ipsrc -> do
-            getdate <- dateHelper mgetdate
-            return $ apacheMiddleware callback ipsrc getdate
-        Detailed useColors -> detailedMiddleware callback useColors
+            getdate <- getDateGetter flusher
+            apache <- initLogger ipsrc (LogCallback callback (return ())) getdate
+            return $ apacheMiddleware apache
+        Detailed useColors -> detailedMiddleware
+                                  (\str -> callback str >> flusher)
+                                  useColors
         CustomOutputFormat formatter -> do
-            getdate <- dateHelper mgetdate
+            getdate <- getDateGetter flusher
             return $ customMiddleware callback getdate formatter
-  where
-    fromLogger l = (loggerPutStr l, Just $ loggerDate l)
-    dateHelper mgetdate = do
-        case mgetdate of
-            Just x -> return x
-#if MIN_VERSION_fast_logger(0, 3, 0)
-            Nothing -> do
-                (getter,_) <- ondemandDateCacher zonedDateCacheConf
-                return getter
-#else
-            Nothing -> fmap getDate dateInit
-#endif
 
-apacheMiddleware :: Callback -> IPAddrSource -> IO ZonedDate -> Middleware
-apacheMiddleware cb ipsrc getdate = customMiddleware cb getdate $ apacheFormat ipsrc
+apacheMiddleware :: ApacheLoggerActions -> Middleware
+apacheMiddleware ala app req sendResponse = app req $ \res -> do
+    let msize = lookup "content-length" (responseHeaders res) >>= readInt'
+        readInt' bs =
+            case S8.readInteger bs of
+                Just (i, "") -> Just i
+                _ -> Nothing
+    apacheLogger ala req (responseStatus res) msize
+    sendResponse res
 
 customMiddleware :: Callback -> IO ZonedDate -> OutputFormatter -> Middleware
-customMiddleware cb getdate formatter app req = do
-    res <- app req
+customMiddleware cb getdate formatter app req sendResponse = app req $ \res -> do
     date <- liftIO getdate
     -- We use Nothing for the response size since we generally don't know it
     liftIO $ cb $ formatter date req (responseStatus res) Nothing
-    return res
+    sendResponse res
 
 -- | Production request logger middleware.
 -- Implemented on top of "logCallback", but prints to 'stdout'
@@ -131,11 +127,10 @@ logStdoutDev = unsafePerformIO $ mkRequestLogger def
 -- no black or white which are expected to be existing terminal colors.
 colors0 :: [Color]
 colors0 = [
-    Red 
-  , Green 
-  , Yellow 
-  , Blue 
-  , Magenta 
+    Green
+  , Yellow
+  , Blue
+  , Magenta
   , Cyan
   ]
 
@@ -171,34 +166,51 @@ rotateColors (c:cs) = (cs ++ [c], c)
 detailedMiddleware :: Callback -> Bool -> IO Middleware
 detailedMiddleware cb useColors = do
     getAddColor <-
-        if useColors
-            then do
+        if not useColors then return (return return) else do
                 icolors <- newIORef colors0
                 return $ do
                     color <- liftIO $ atomicModifyIORef icolors rotateColors
                     return $ ansiColor color
-            else return (return return)
     return $ detailedMiddleware' cb getAddColor
-  where
-    ansiColor color bs = [
-        pack $ setSGRCode [SetColor Foreground Vivid color]
-      , bs
-      , pack $ setSGRCode [Reset]
-      ]
+
+ansiColor :: Color -> BS.ByteString -> [BS.ByteString]
+ansiColor color bs = [
+    pack $ setSGRCode [SetColor Foreground Vivid color]
+  , bs
+  , pack $ setSGRCode [Reset]
+  ]
 
 detailedMiddleware' :: Callback
-                    -> (C.ResourceT IO (BS.ByteString -> [BS.ByteString]))
+                    -> IO (BS.ByteString -> [BS.ByteString])
                     -> Middleware
-detailedMiddleware' cb getAddColor app req = do
+detailedMiddleware' cb getAddColor app req sendResponse = do
     let mlen = lookup "content-length" (requestHeaders req) >>= readInt
     (req', body) <-
         case mlen of
             -- log the request body if it is small
             Just len | len <= 2048 -> do
-                 body <- requestBody req C.$$ CL.consume
+                 let loop front = do
+                        bs <- requestBody req
+                        if S8.null bs
+                            then return $ front []
+                            else loop $ front . (bs:)
+                 body <- loop id
                  -- logging the body here consumes it, so fill it back up
                  -- obviously not efficient, but this is the development logger
-                 let req' = req { requestBody = CL.sourceList body }
+                 --
+                 -- Note: previously, we simply used CL.sourceList. However,
+                 -- that meant that you could read the request body in twice.
+                 -- While that in itself is not a problem, the issue is that,
+                 -- in production, you wouldn't be able to do this, and
+                 -- therefore some bugs wouldn't show up during testing. This
+                 -- implementation ensures that each chunk is only returned
+                 -- once.
+                 ichunks <- newIORef body
+                 let rbody = atomicModifyIORef ichunks $ \chunks ->
+                        case chunks of
+                            [] -> ([], S8.empty)
+                            x:y -> (y, x)
+                 let req' = req { requestBody = rbody }
                  return (req', body)
             _ -> return (req, [])
 
@@ -210,33 +222,37 @@ detailedMiddleware' cb getAddColor app req = do
     let getParams = map emptyGetParam $ queryString req
 
     addColor <- getAddColor
+    let accept = fromMaybe "" $ lookup "Accept" $ requestHeaders req
 
     -- log the request immediately.
-    liftIO $ cb $ map LB $ addColor (requestMethod req) ++
+    liftIO $ cb $ mconcat $ map toLogStr $ addColor (requestMethod req) ++
         [ " "
         , rawPathInfo req
-        , "\n"
-        , "Accept: "
-        , fromMaybe "" $ lookup "Accept" $ requestHeaders req
+        , " :: "
+        , accept
         , paramsToBS  "GET " getParams
         , paramsToBS "POST " postParams
         , "\n"
         ]
 
-    rsp <- app req'
+    app req' $ \rsp -> do
+        let isRaw =
+                case rsp of
+                    ResponseRaw{} -> True
+                    _ -> False
 
-    -- log the status of the response
-    -- this is color coordinated with the request logging
-    -- also includes the request path to connect it to the request
-    liftIO $ cb $ map LB $ addColor "Status: " ++ [
-          statusBS rsp
-        , " "
-        , msgBS rsp
-        , ". "
-        , rawPathInfo req -- if you need help matching the 2 logging statements
-        , "\n"
-      ]
-    return rsp
+        -- log the status of the response
+        -- this is color coordinated with the request logging
+        -- also includes the request path to connect it to the request
+        unless isRaw $ cb $ mconcat $ map toLogStr $
+            addColor "Status: " ++ statusBS rsp ++
+            [ " "
+            , msgBS rsp
+            , ". "
+            , rawPathInfo req -- if you need help matching the 2 logging statements
+            , "\n"
+            ]
+        sendResponse rsp
   where
     paramsToBS prefix params =
       if null params then ""
@@ -245,7 +261,13 @@ detailedMiddleware' cb getAddColor app req = do
     allPostParams body =
         case getRequestBodyType req of
             Nothing -> return ([], [])
-            Just rbt -> C.runResourceT $ CL.sourceList body C.$$ sinkRequestBody lbsBackEnd rbt
+            Just rbt -> do
+                ichunks <- newIORef body
+                let rbody = atomicModifyIORef ichunks $ \chunks ->
+                        case chunks of
+                            [] -> ([], S8.empty)
+                            x:y -> (y, x)
+                sinkRequestBody lbsBackEnd rbt rbody
 
     emptyGetParam :: (BS.ByteString, Maybe BS.ByteString) -> (BS.ByteString, BS.ByteString)
     emptyGetParam (k, Just v) = (k,v)
@@ -260,8 +282,12 @@ detailedMiddleware' cb getAddColor app req = do
             (i, _):_ -> Just (i :: Int)
             [] -> Nothing
 
-statusBS :: Response -> BS.ByteString
-statusBS = pack . show . statusCode . responseStatus
+statusBS :: Response -> [BS.ByteString]
+statusBS rsp =
+    if status >= 400 then ansiColor Red bs else [bs]
+  where
+    bs = pack $ show status
+    status = statusCode $ responseStatus rsp
 
 msgBS :: Response -> BS.ByteString
 msgBS = statusMessage . responseStatus
