@@ -12,7 +12,6 @@ module Network.Wai.Parse
     , RequestBodyType (..)
     , getRequestBodyType
     , sinkRequestBody
-    , conduitRequestBody
     , BackEnd
     , lbsBackEnd
     , tempFileBackEnd
@@ -38,24 +37,14 @@ import qualified Data.ByteString.Char8 as S8
 import Data.Word (Word8)
 import Data.Maybe (fromMaybe)
 import Data.List (sortBy)
-import Data.Function (on)
+import Data.Function (on, fix)
 import System.Directory (removeFile, getTemporaryDirectory)
 import System.IO (hClose, openBinaryTempFile)
 import Network.Wai
-import Data.Conduit
-import Data.Conduit.Internal (sinkToPipe)
-import qualified Data.Conduit.List as CL
-import qualified Data.Conduit.Binary as CB
-import Control.Monad.IO.Class (liftIO)
 import qualified Network.HTTP.Types as H
-import Data.Either (partitionEithers)
 import Control.Monad (when, unless)
-import Control.Monad.Trans.Class (lift)
-import Control.Monad.Trans.Resource (allocate, release, register)
-#if MIN_VERSION_conduit(1, 0, 0)
-import Data.Conduit.Internal (Pipe (NeedInput, HaveOutput), (>+>), withUpstream, Sink (..), injectLeftovers, ConduitM (..))
-import Data.Void (Void)
-#endif
+import Control.Monad.Trans.Resource (allocate, release, register, InternalState, runInternalState)
+import Data.IORef
 
 breakDiscard :: Word8 -> S.ByteString -> (S.ByteString, S.ByteString)
 breakDiscard w s =
@@ -87,27 +76,43 @@ parseHttpAccept = map fst
                 _ -> 1.0
 
 -- | Store uploaded files in memory
-lbsBackEnd :: Monad m => ignored1 -> ignored2 -> Sink S.ByteString m L.ByteString
-lbsBackEnd _ _ = fmap L.fromChunks CL.consume
+lbsBackEnd :: Monad m => ignored1 -> ignored2 -> m S.ByteString -> m L.ByteString
+lbsBackEnd _ _ popper =
+    loop id
+  where
+    loop front = do
+        bs <- popper
+        if S.null bs
+            then return $ L.fromChunks $ front []
+            else loop $ front . (bs:)
 
 -- | Save uploaded files on disk as temporary files
-tempFileBackEnd :: MonadResource m => ignored1 -> ignored2 -> Sink S.ByteString m FilePath
+--
+-- Note: starting with version 2.0, removal of temp files is registered with
+-- the provided @InternalState@. It is the responsibility of the caller to
+-- ensure that this @InternalState@ gets cleaned up.
+tempFileBackEnd :: InternalState -> ignored1 -> ignored2 -> IO S.ByteString -> IO FilePath
 tempFileBackEnd = tempFileBackEndOpts getTemporaryDirectory "webenc.buf"
 
 -- | Same as 'tempFileSink', but use configurable temp folders and patterns.
-tempFileBackEndOpts :: MonadResource m
-                    => IO FilePath -- ^ get temporary directory
+tempFileBackEndOpts :: IO FilePath -- ^ get temporary directory
                     -> String -- ^ filename pattern
+                    -> InternalState
                     -> ignored1
                     -> ignored2
-                    -> Sink S.ByteString m FilePath
-tempFileBackEndOpts getTmpDir pattern _ _ = do
-    (key, (fp, h)) <- lift $ allocate (do
+                    -> IO S.ByteString
+                    -> IO FilePath
+tempFileBackEndOpts getTmpDir pattern internalState _ _ popper = do
+    (key, (fp, h)) <- flip runInternalState internalState $ allocate (do
         tempDir <- getTmpDir
         openBinaryTempFile tempDir pattern) (\(_, h) -> hClose h)
-    _ <- lift $ register $ removeFile fp
-    CB.sinkHandle h
-    lift $ release key
+    _ <- runInternalState (register $ removeFile fp) internalState
+    fix $ \loop -> do
+        bs <- popper
+        unless (S.null bs) $ do
+            S.hPut h bs
+            loop
+    release key
     return fp
 
 -- | Information on an uploaded file.
@@ -124,11 +129,12 @@ type Param = (S.ByteString, S.ByteString)
 -- | Post parameter name and associated file information.
 type File y = (S.ByteString, FileInfo y)
 
--- | A file uploading backend. Takes the parameter name, file name, and content
--- type, and returns a `Sink` for storing the contents.
+-- | A file uploading backend. Takes the parameter name, file name, and a
+-- stream of data.
 type BackEnd a = S.ByteString -- ^ parameter name
               -> FileInfo ()
-              -> Sink S.ByteString (ResourceT IO) a
+              -> IO S.ByteString
+              -> IO a
 
 data RequestBodyType = UrlEncoded | Multipart S.ByteString
 
@@ -167,76 +173,103 @@ parseContentType a = do
 
 parseRequestBody :: BackEnd y
                  -> Request
-                 -> ResourceT IO ([Param], [File y])
+                 -> IO ([Param], [File y])
 parseRequestBody s r =
     case getRequestBodyType r of
         Nothing -> return ([], [])
-        Just rbt -> fmap partitionEithers $ requestBody r $$ conduitRequestBody s rbt =$ CL.consume
+        Just rbt -> sinkRequestBody s rbt (requestBody r)
 
 sinkRequestBody :: BackEnd y
                 -> RequestBodyType
-                -> Sink S.ByteString (ResourceT IO) ([Param], [File y])
-sinkRequestBody s r = fmap partitionEithers $ conduitRequestBody s r =$ CL.consume
+                -> IO S.ByteString
+                -> IO ([Param], [File y])
+sinkRequestBody s r body = do
+    ref <- newIORef (id, id)
+    let add x = atomicModifyIORef ref $ \(y, z) ->
+            case x of
+                Left y' -> ((y . (y':), z), ())
+                Right z' -> ((y, z . (z':)), ())
+    conduitRequestBody s r body add
+    (x, y) <- readIORef ref
+    return (x [], y [])
 
 conduitRequestBody :: BackEnd y
                    -> RequestBodyType
-                   -> Conduit S.ByteString (ResourceT IO) (Either Param (File y))
-conduitRequestBody _ UrlEncoded = do
+                   -> IO S.ByteString
+                   -> (Either Param (File y) -> IO ())
+                   -> IO ()
+conduitRequestBody _ UrlEncoded rbody add = do
     -- NOTE: in general, url-encoded data will be in a single chunk.
     -- Therefore, I'm optimizing for the usual case by sticking with
     -- strict byte strings here.
-    bs <- CL.consume
-    mapM_ yield $ map Left $ H.parseSimpleQuery $ S.concat bs
-conduitRequestBody backend (Multipart bound) =
-    parsePieces backend $ S8.pack "--" `S.append` bound
+    let loop front = do
+            bs <- rbody
+            if S.null bs
+                then return $ S.concat $ front []
+                else loop $ front . (bs:)
+    bs <- loop id
+    mapM_ (add . Left) $ H.parseSimpleQuery bs
+conduitRequestBody backend (Multipart bound) rbody add =
+    parsePieces backend (S8.pack "--" `S.append` bound) rbody add
 
-#if MIN_VERSION_conduit(1, 0, 0)
-takeLine :: Monad m => Consumer S.ByteString m (Maybe S.ByteString)
-#else
-takeLine :: Monad m => Pipe S.ByteString S.ByteString o u m (Maybe S.ByteString)
-#endif
-takeLine =
+takeLine :: Source -> IO (Maybe S.ByteString)
+takeLine src =
     go id
   where
-    go front = await >>= maybe (close front) (push front)
+    go front = do
+        bs <- readSource src
+        if S.null bs
+            then close front
+            else push front bs
 
-    close front = leftover (front S.empty) >> return Nothing
+    close front = leftover src (front S.empty) >> return Nothing
     push front bs = do
         let (x, y) = S.break (== 10) $ front bs -- LF
          in if S.null y
                 then go $ S.append x
                 else do
-                    when (S.length y > 1) $ leftover $ S.drop 1 y
+                    when (S.length y > 1) $ leftover src $ S.drop 1 y
                     return $ Just $ killCR x
 
-#if MIN_VERSION_conduit(1, 0, 0)
-takeLines :: Consumer S.ByteString (ResourceT IO) [S.ByteString]
-#else
-takeLines :: Pipe S.ByteString S.ByteString o u (ResourceT IO) [S.ByteString]
-#endif
-takeLines = do
-    res <- takeLine
+takeLines :: Source -> IO [S.ByteString]
+takeLines src = do
+    res <- takeLine src
     case res of
         Nothing -> return []
         Just l
             | S.null l -> return []
             | otherwise -> do
-                ls <- takeLines
+                ls <- takeLines src
                 return $ l : ls
 
+data Source = Source (IO S.ByteString) (IORef S.ByteString)
+
+mkSource :: IO S.ByteString -> IO Source
+mkSource f = do
+    ref <- newIORef S.empty
+    return $ Source f ref
+
+readSource :: Source -> IO S.ByteString
+readSource (Source f ref) = do
+    bs <- atomicModifyIORef ref $ \bs -> (S.empty, bs)
+    if S.null bs
+        then f
+        else return bs
+
+leftover :: Source -> S.ByteString -> IO ()
+leftover (Source _ ref) bs = writeIORef ref bs
+
 parsePieces :: BackEnd y
             -> S.ByteString
-#if MIN_VERSION_conduit(1, 0, 0)
-            -> ConduitM S.ByteString (Either Param (File y)) (ResourceT IO) ()
-#else
-            -> Pipe S.ByteString S.ByteString (Either Param (File y)) u (ResourceT IO) ()
-#endif
-parsePieces sink bound =
-    loop
+            -> IO S.ByteString
+            -> (Either Param (File y) -> IO ())
+            -> IO ()
+parsePieces sink bound rbody add =
+    mkSource rbody >>= loop
   where
-    loop = do
-        _boundLine <- takeLine
-        res' <- takeLines
+    loop src = do
+        _boundLine <- takeLine src
+        res' <- takeLines src
         unless (null res') $ do
             let ls' = map parsePair res'
             let x = do
@@ -249,23 +282,23 @@ parsePieces sink bound =
                 Just (mct, name, Just filename) -> do
                     let ct = fromMaybe "application/octet-stream" mct
                         fi0 = FileInfo filename ct ()
-                    (wasFound, y) <- sinkTillBound' bound name fi0 sink
-                    yield $ Right (name, fi0 { fileContent = y })
-                    when wasFound loop
+                    (wasFound, y) <- sinkTillBound' bound name fi0 sink src
+                    add $ Right (name, fi0 { fileContent = y })
+                    when wasFound (loop src)
                 Just (_ct, name, Nothing) -> do
                     let seed = id
                     let iter front bs = return $ front . (:) bs
-                    (wasFound, front) <- sinkTillBound bound iter seed
+                    (wasFound, front) <- sinkTillBound bound iter seed src
                     let bs = S.concat $ front []
                     let x' = (name, bs)
-                    yield $ Left x'
-                    when wasFound loop
+                    add $ Left x'
+                    when wasFound (loop src)
                 _ -> do
                     -- ignore this part
                     let seed = ()
                         iter () _ = return ()
-                    (wasFound, ()) <- sinkTillBound bound iter seed
-                    when wasFound loop
+                    (wasFound, ()) <- sinkTillBound bound iter seed src
+                    when wasFound (loop src)
       where
         contDisp = S8.pack "Content-Disposition"
         contType = S8.pack "Content-Type"
@@ -305,90 +338,78 @@ sinkTillBound' :: S.ByteString
                -> S.ByteString
                -> FileInfo ()
                -> BackEnd y
-#if MIN_VERSION_conduit(1, 0, 0)
-               -> ConduitM S.ByteString o (ResourceT IO) (Bool, y)
-#else
-               -> Pipe S.ByteString S.ByteString o u (ResourceT IO) (Bool, y)
-#endif
-sinkTillBound' bound name fi sink =
-#if MIN_VERSION_conduit(1, 0, 0)
-    ConduitM $ anyOutput $
-#endif
-    conduitTillBound bound >+> withUpstream (fix $ sink name fi)
-  where
-#if MIN_VERSION_conduit(1, 0, 0)
-    fix :: Sink S8.ByteString (ResourceT IO) y -> Pipe Void S8.ByteString Void Bool (ResourceT IO) y
-    fix (ConduitM p) = ignoreTerm >+> injectLeftovers p
-    ignoreTerm = await' >>= maybe (return ()) (\x -> yield' x >> ignoreTerm)
-    await' = NeedInput (return . Just) (const $ return Nothing)
-    yield' = HaveOutput (return ()) (return ())
+               -> Source
+               -> IO (Bool, y)
+sinkTillBound' bound name fi sink src = do
+    (next, final) <- wrapTillBound bound src
+    y <- sink name fi next
+    b <- final
+    return (b, y)
 
-    anyOutput p = p >+> dropInput
-    dropInput = NeedInput (const dropInput) return
-#else
-    fix = sinkToPipe
-#endif
-
-conduitTillBound :: Monad m
-                 => S.ByteString -- bound
-#if MIN_VERSION_conduit(1, 0, 0)
-                 -> Pipe S.ByteString S.ByteString S.ByteString () m Bool
-#else
-                 -> Pipe S.ByteString S.ByteString S.ByteString u m Bool
-#endif
-conduitTillBound bound =
-#if MIN_VERSION_conduit(1, 0, 0)
-    unConduitM $
-#endif
-    go id
+data WTB = WTBWorking (S.ByteString -> S.ByteString)
+         | WTBDone Bool
+wrapTillBound :: S.ByteString -- ^ bound
+              -> Source
+              -> IO (IO S.ByteString, IO Bool) -- ^ Bool indicates if the bound was found
+wrapTillBound bound src = do
+    ref <- newIORef $ WTBWorking id
+    return (go ref, final ref)
   where
-    go front = await >>= maybe (close front) (push front)
-    close front = do
-        let bs = front S.empty
-        unless (S.null bs) $ yield bs
-        return False
-    push front bs' = do
-        let bs = front bs'
-        case findBound bound bs of
-            FoundBound before after -> do
-                let before' = killCRLF before
-                yield before'
-                leftover after
-                return True
-            NoBound -> do
-                -- don't emit newlines, in case it's part of a bound
-                let (toEmit, front') =
-                        if not (S8.null bs) && S8.last bs `elem` "\r\n"
-                            then let (x, y) = S.splitAt (S.length bs - 2) bs
-                                  in (x, S.append y)
-                            else (bs, id)
-                yield toEmit
-                go front'
-            PartialBound -> go $ S.append bs
+    final ref = do
+        x <- readIORef ref
+        case x of
+            WTBWorking _ -> error "wrapTillBound did not finish"
+            WTBDone y -> return y
+
+    go ref = do
+        state <- readIORef ref
+        case state of
+            WTBDone _ -> return S.empty
+            WTBWorking front -> do
+                bs <- readSource src
+                if S.null bs
+                    then do
+                        writeIORef ref $ WTBDone False
+                        return $ front bs
+                    else push $ front bs
+      where
+        push bs =
+            case findBound bound bs of
+                FoundBound before after -> do
+                    let before' = killCRLF before
+                    leftover src after
+                    writeIORef ref $ WTBDone True
+                    return before'
+                NoBound -> do
+                    -- don't emit newlines, in case it's part of a bound
+                    let (toEmit, front') =
+                            if not (S8.null bs) && S8.last bs `elem` "\r\n"
+                                then let (x, y) = S.splitAt (S.length bs - 2) bs
+                                      in (x, S.append y)
+                                else (bs, id)
+                    writeIORef ref $ WTBWorking front'
+                    if S.null toEmit
+                        then go ref
+                        else return toEmit
+                PartialBound -> do
+                    writeIORef ref $ WTBWorking $ S.append bs
+                    go ref
 
 sinkTillBound :: S.ByteString
               -> (x -> S.ByteString -> IO x)
               -> x
-#if MIN_VERSION_conduit(1, 0, 0)
-              -> Consumer S.ByteString (ResourceT IO) (Bool, x)
-#else
-              -> Pipe S.ByteString S.ByteString o u (ResourceT IO) (Bool, x)
-#endif
-sinkTillBound bound iter seed0 =
-#if MIN_VERSION_conduit(1, 0, 0)
-    ConduitM $
-#endif
-    (conduitTillBound bound >+> (withUpstream $ ij $ CL.foldM iter' seed0))
-  where
-    iter' a b = liftIO $ iter a b
-#if MIN_VERSION_conduit(1, 0, 0)
-    ij (ConduitM p) = ignoreTerm >+> injectLeftovers p
-    ignoreTerm = await' >>= maybe (return ()) (\x -> yield' x >> ignoreTerm)
-    await' = NeedInput (return . Just) (const $ return Nothing)
-    yield' = HaveOutput (return ()) (return ())
-#else
-    ij = id
-#endif
+              -> Source
+              -> IO (Bool, x)
+sinkTillBound bound iter seed0 src = do
+    (next, final) <- wrapTillBound bound src
+    let loop seed = do
+            bs <- next
+            if S.null bs
+                then return seed
+                else iter seed bs >>= loop
+    seed <- loop seed0
+    b <- final
+    return (b, seed)
 
 parseAttrs :: S.ByteString -> [(S.ByteString, S.ByteString)]
 parseAttrs = map go . S.split 59 -- semicolon
