@@ -3,10 +3,12 @@ module WaiExtraTest (specs) where
 
 import Test.Hspec
 import Test.HUnit hiding (Test)
+import Data.Monoid (mappend, mempty)
 
 import Network.Wai
 import Network.Wai.Test
 import Network.Wai.Parse
+import Network.Wai.UrlMap
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Char8 as S8
 import qualified Data.ByteString.Lazy.Char8 as L8
@@ -15,6 +17,8 @@ import qualified Data.Text.Lazy as T
 import qualified Data.Text as TS
 import qualified Data.Text.Encoding as TE
 import Control.Arrow
+import Control.Applicative
+import Control.Monad.Trans.Resource (withInternalState, runResourceT)
 
 import Network.Wai.Middleware.Jsonp
 import Network.Wai.Middleware.Gzip
@@ -26,18 +30,19 @@ import Network.Wai.Middleware.AcceptOverride
 import Network.Wai.Middleware.RequestLogger
 import Codec.Compression.GZip (decompress)
 
-import qualified Data.Conduit as C
-import qualified Data.Conduit.List as CL
-import Data.Conduit.Binary (sourceFile)
 import Control.Monad.IO.Class (liftIO)
 import Data.Maybe (fromMaybe)
 import Network.HTTP.Types (parseSimpleQuery, status200)
 import System.Log.FastLogger
+import System.IO (withFile, IOMode (ReadMode))
 
 import qualified Data.IORef as I
 
 specs :: Spec
 specs = do
+  describe "Network.Wai.UrlMap" $ do
+    mapM_ (uncurry it) casesUrlMap
+
   describe "Network.Wai.Parse" $ do
     describe "parseContentType" $ do
         let go (x, y, z) = it (TS.unpack $ TE.decodeUtf8 x) $ parseContentType x `shouldBe` (y, z)
@@ -63,6 +68,7 @@ specs = do
     it "jsonp" caseJsonp
     it "gzip" caseGzip
     it "gzip not for MSIE" caseGzipMSIE
+    it "gzip bypass when precompressed" caseGzipBypassPre
     it "defaultCheckMime" caseDefaultCheckMime
     it "vhost" caseVhost
     it "autohead" caseAutohead
@@ -113,17 +119,23 @@ caseParseHttpAccept = do
         expected = ["text/html;charset=utf-8", "text/x-c", "text/x-dvi", "text/*", "text/plain"]
     expected @=? parseHttpAccept input
 
-parseRequestBody' :: BackEnd L.ByteString
+parseRequestBody' :: BackEnd file
                   -> SRequest
-                  -> C.ResourceT IO ([(S.ByteString, S.ByteString)], [(S.ByteString, FileInfo L.ByteString)])
+                  -> IO ([(S.ByteString, S.ByteString)], [(S.ByteString, FileInfo file)])
 parseRequestBody' sink (SRequest req bod) =
     case getRequestBodyType req of
         Nothing -> return ([], [])
-        Just rbt -> CL.sourceList (L.toChunks bod) C.$$ sinkRequestBody sink rbt
+        Just rbt -> do
+            ref <- I.newIORef $ L.toChunks bod
+            let rb = I.atomicModifyIORef ref $ \chunks ->
+                        case chunks of
+                            [] -> ([], S.empty)
+                            x:y -> (y, x)
+            sinkRequestBody sink rbt rb
 
 caseParseRequestBody :: Assertion
 caseParseRequestBody =
-    C.runResourceT t
+    t
   where
     content2 = S8.pack $
         "--AaB03x\n" ++
@@ -187,7 +199,7 @@ caseParseRequestBody =
 
 caseMultipartPlus :: Assertion
 caseMultipartPlus = do
-    result <- C.runResourceT $ parseRequestBody' lbsBackEnd $ toRequest ctype content
+    result <- parseRequestBody' lbsBackEnd $ toRequest ctype content
     liftIO $ result @?= ([("email", "has+plus")], [])
   where
     content = S8.pack $
@@ -200,7 +212,7 @@ caseMultipartPlus = do
 
 caseMultipartAttrs :: Assertion
 caseMultipartAttrs = do
-    result <- C.runResourceT $ parseRequestBody' lbsBackEnd $ toRequest ctype content
+    result <- parseRequestBody' lbsBackEnd $ toRequest ctype content
     liftIO $ result @?= ([("email", "has+plus")], [])
   where
     content = S8.pack $
@@ -213,7 +225,8 @@ caseMultipartAttrs = do
 
 caseUrlEncPlus :: Assertion
 caseUrlEncPlus = do
-    result <- C.runResourceT $ parseRequestBody' lbsBackEnd $ toRequest ctype content
+    result <- runResourceT $ withInternalState $ \state ->
+              parseRequestBody' (tempFileBackEnd state) $ toRequest ctype content
     liftIO $ result @?= ([("email", "has+plus")], [])
   where
     content = S8.pack $ "email=has%2Bplus"
@@ -285,7 +298,7 @@ caseTakeLine = do
 -}
 
 jsonpApp :: Application
-jsonpApp = jsonp $ const $ return $ responseLBS
+jsonpApp = jsonp $ \_ f -> f $ responseLBS
     status200
     [("Content-Type", "application/json")]
     "{\"foo\":\"bar\"}"
@@ -314,10 +327,18 @@ caseJsonp = flip runSession jsonpApp $ do
     assertBody "{\"foo\":\"bar\"}" sres3
 
 gzipApp :: Application
-gzipApp = gzip def $ const $ return $ responseLBS status200
+gzipApp = gzip def $ \_ f -> f $ responseLBS status200
     [("Content-Type", "text/plain")]
     "test"
 
+-- Lie a little and don't compress the body.  This way we test
+-- that the compression is skipped based on the presence of
+-- the Content-Encoding header.
+gzipPrecompressedApp :: Application
+gzipPrecompressedApp = gzip def $ \_ f -> f $ responseLBS status200
+    [("Content-Type", "text/plain"), ("Content-Encoding", "gzip")]
+    "test"
+
 caseGzip :: Assertion
 caseGzip = flip runSession gzipApp $ do
     sres1 <- request defaultRequest
@@ -353,28 +374,36 @@ caseGzipMSIE = flip runSession gzipApp $ do
     assertNoHeader "Content-Encoding" sres1
     liftIO $ simpleBody sres1 @?= "test"
 
+caseGzipBypassPre :: Assertion
+caseGzipBypassPre = flip runSession gzipPrecompressedApp $ do
+    sres1 <- request defaultRequest
+                { requestHeaders = [("Accept-Encoding", "gzip")]
+                }
+    assertHeader "Content-Encoding" "gzip" sres1
+    assertBody "test" sres1 -- the body is not actually compressed
+
 vhostApp1, vhostApp2, vhostApp :: Application
-vhostApp1 = const $ return $ responseLBS status200 [] "app1"
-vhostApp2 = const $ return $ responseLBS status200 [] "app2"
+vhostApp1 _ f = f $ responseLBS status200 [] "app1"
+vhostApp2 _ f = f $ responseLBS status200 [] "app2"
 vhostApp = vhost
-    [ ((== "foo.com") . serverName, vhostApp1)
+    [ ((== Just "foo.com") . lookup "host" . requestHeaders, vhostApp1)
     ]
     vhostApp2
 
 caseVhost :: Assertion
 caseVhost = flip runSession vhostApp $ do
     sres1 <- request defaultRequest
-                { serverName = "foo.com"
+                { requestHeaders = [("Host", "foo.com")]
                 }
     assertBody "app1" sres1
 
     sres2 <- request defaultRequest
-                { serverName = "bar.com"
+                { requestHeaders = [("Host", "bar.com")]
                 }
     assertBody "app2" sres2
 
 autoheadApp :: Application
-autoheadApp = autohead $ const $ return $ responseLBS status200
+autoheadApp = autohead $ \_ f -> f $ responseLBS status200
     [("Foo", "Bar")] "body"
 
 caseAutohead :: Assertion
@@ -392,7 +421,7 @@ caseAutohead = flip runSession autoheadApp $ do
     assertBody "" sres2
 
 moApp :: Application
-moApp = methodOverride $ \req -> return $ responseLBS status200
+moApp = methodOverride $ \req f -> f $ responseLBS status200
     [("Method", requestMethod req)] ""
 
 caseMethodOverride :: Assertion
@@ -416,7 +445,7 @@ caseMethodOverride = flip runSession moApp $ do
     assertHeader "Method" "PUT" sres3
 
 mopApp :: Application
-mopApp = methodOverridePost $ \req -> return $ responseLBS status200 [("Method", requestMethod req)] ""
+mopApp = methodOverridePost $ \req f -> f $ responseLBS status200 [("Method", requestMethod req)] ""
 
 caseMethodOverridePost :: Assertion
 caseMethodOverridePost = flip runSession mopApp $ do
@@ -442,7 +471,7 @@ caseMethodOverridePost = flip runSession mopApp $ do
     assertHeader "Method" "POST" sres4
 
 aoApp :: Application
-aoApp = acceptOverride $ \req -> return $ responseLBS status200
+aoApp = acceptOverride $ \req f -> f $ responseLBS status200
     [("Accept", fromMaybe "" $ lookup "Accept" $ requestHeaders req)] ""
 
 caseAcceptOverride :: Assertion
@@ -493,8 +522,8 @@ dalvikHelper includeLength = do
     (params, files) <-
         case getRequestBodyType request' of
             Nothing -> return ([], [])
-            Just rbt -> C.runResourceT $ sourceFile "test/requests/dalvik-request"
-                       C.$$ sinkRequestBody lbsBackEnd rbt
+            Just rbt -> withFile "test/requests/dalvik-request" ReadMode $ \h ->
+                sinkRequestBody lbsBackEnd rbt $ S.hGetSome h 2048
     lookup "scannedTime" params @?= Just "1.298590056748E9"
     lookup "geoLong" params @?= Just "0"
     lookup "geoLat" params @?= Just "0"
@@ -519,25 +548,83 @@ caseDebugRequestBody = do
   where
     params = [("foo", "bar"), ("baz", "bin")]
     -- FIXME change back once we include post parameter output in logging postOutput = T.pack $ "POST \nAccept: \nPOST " ++ (show params)
-    postOutput = T.pack $ "POST /\nAccept: \nStatus: 200 OK. /\n"
-    getOutput params' = T.pack $ "GET /location\nAccept: \nGET " ++ show params' ++ "\nStatus: 200 OK. /location\n"
+    postOutput = T.pack $ "POST / :: \nStatus: 200 OK. /\n"
+    getOutput params' = T.pack $ "GET /location :: \nGET " ++ show params' ++ "\nStatus: 200 OK. /location\n"
 
-    debugApp output' req = do
-        iactual <- liftIO $ I.newIORef []
-        middleware <- liftIO $ mkRequestLogger def
-            { destination = Callback $ \strs -> I.modifyIORef iactual $ (++ strs)
+    debugApp output' req send = do
+        iactual <- I.newIORef mempty
+        middleware <- mkRequestLogger def
+            { destination = Callback $ \strs -> I.modifyIORef iactual $ (`mappend` strs)
             , outputFormat = Detailed False
             }
-        res <- middleware (\_req -> return $ responseLBS status200 [ ] "") req
-        actual <- liftIO $ I.readIORef iactual
-        liftIO $ assertEqual "debug" output $ logsToBs actual
+        res <- middleware (\_req f -> f $ responseLBS status200 [ ] "") req send
+        actual <- I.readIORef iactual
+        assertEqual "debug" output $ logToBs actual
         return res
       where
         output = TE.encodeUtf8 $ T.toStrict output'
-        logsToBs = S.concat . map logToBs
 
-        logToBs (LB bs) = bs
-        logToBs (LS s) = S8.pack s
+        logToBs = fromLogStr
 
     {-debugApp = debug $ \req -> do-}
         {-return $ responseLBS status200 [ ] ""-}
+
+urlMapTestApp :: Application
+urlMapTestApp = mapUrls $
+        mount "bugs"     bugsApp
+    <|> mount "helpdesk" helpdeskApp
+    <|> mount "api"
+            (   mount "v1" apiV1
+            <|> mount "v2" apiV2
+            )
+    <|> mountRoot mainApp
+
+  where
+  trivialApp :: S.ByteString -> Application
+  trivialApp name req f =
+    f $
+      responseLBS
+        status200
+        [ ("content-type", "text/plain")
+        , ("X-pathInfo",    S8.pack . show . pathInfo $ req)
+        , ("X-rawPathInfo", rawPathInfo req)
+        , ("X-appName",     name)
+        ]
+        ""
+
+  bugsApp     = trivialApp "bugs"
+  helpdeskApp = trivialApp "helpdesk"
+  apiV1       = trivialApp "apiv1"
+  apiV2       = trivialApp "apiv2"
+  mainApp     = trivialApp "main"
+
+casesUrlMap :: [(String, Assertion)]
+casesUrlMap = [pair1, pair2, pair3, pair4]
+  where
+  makePair name session = (name, runSession session urlMapTestApp)
+  get reqPath = request $ setPath defaultRequest reqPath
+  s = S8.pack . show :: [TS.Text] -> S.ByteString
+
+  pair1 = makePair "should mount root" $ do
+    res1 <- get "/"
+    assertStatus 200 res1
+    assertHeader "X-rawPathInfo" "/"    res1
+    assertHeader "X-pathInfo"    (s []) res1
+    assertHeader "X-appName"     "main" res1
+
+  pair2 = makePair "should mount apps" $ do
+    res2 <- get "/bugs"
+    assertStatus 200 res2
+    assertHeader "X-rawPathInfo" "/"    res2
+    assertHeader "X-pathInfo"    (s []) res2
+    assertHeader "X-appName"     "bugs" res2
+
+  pair3 = makePair "should preserve extra path info" $ do
+    res3 <- get "/helpdesk/issues/11"
+    assertStatus 200 res3
+    assertHeader "X-rawPathInfo" "/issues/11"         res3
+    assertHeader "X-pathInfo"    (s ["issues", "11"]) res3
+
+  pair4 = makePair "should 404 if none match" $ do
+    res4 <- get "/api/v3"
+    assertStatus 404 res4
